/*****************************************************************************/
/*
                                   CGI.c

Provide generalized CGI scripting support.

Creates a buffer containing either, the DCL commands to create CGI variables at
the VMS command line (e.g. for standard CGI scripts), or a stream of
'name=value' pairs for the CGI variables (e.g. for CGIplus scripts). 

The buffer is dynamically allocated and contains a series of sequential records
comprising a word (16 bits) with the length of a following (varying)
null-terminated string (including the null character).  The end of the series
is indicated by a zero length record.

This buffer is scanned from from first to last and the strings passed to the
script 'input' stream as appropriate.


CGI VARIABLES
-------------
Most of these CGI variable names are those supported by the INTERNET-DRAFT
authored by D.Robinson (drtr@ast.cam.ac.uk), 8 January 1996, plus some
"convenience" variables, breaking the query string into its components (KEY_,
FORM_, etc.) The CGI symbols (CERN/VMS-HTTPd-like DCL symbols instead of Unix
environment variables) are created by the SYS$COMMAND stream of each
subprocess before the script DCL procedure is invoked. By default each
variable name is prefixed by "WWW_" (similar to CERN HTTPd), although this can
be modified at the command line when starting the server. CGI variable
(symbol) values are limited in size to 1020 characters. 

     VARIABLE NAME            DESCRIPTION                            "standard"
     -------------            -----------                            ----------

  (if the request is not authorized only AUTH_TYPE will exist)
  o  AUTH_ACCESS ............ "READ" or "READ+WRITE"                 [no]
  o  AUTH_AGENT ............. indicates authentication agent         [no]
  o  AUTH_GROUP ............. path authorization group               [no]
  o  AUTH_PASSWORD .......... only if "EXTERNAL" realm               [no]
  o  AUTH_REALM ............. authentication realm                   [no]
  o  AUTH_REALM_DESCRIPTION . realm description                      [no]
  o  AUTH_TYPE .............. "BASIC" or "DIGEST" (or empty)         [yes]
  o  AUTH_USER .............. details of user                        [no]

  o  CONTENT_LENGTH ......... "Content-Length:" from header          [yes]
  o  CONTENT_TYPE ........... "Content-Type:" from header            [yes]

  o  FORM_field ............. query "&" separated form elements      [no]

  o  GATEWAY_INTERFACE ...... "CGI/1.1"                              [yes]

  o  HTTP_ACCEPT ............ list of browser-accepted content types [yes]
  O  HTTP_ACCEPT_ENCODING ... list of browser-accepted encodings     [yes]
  o  HTTP_ACCEPT_CHARSET .... list of browser-accepted character set [yes]
  o  HTTP_ACCEPT_LANGUAGE ... list of browser-accepted languages     [yes]
  o  HTTP_AUTHORIZATION ..... only if "EXTERNAL" realm               [yes]
  o  HTTP_COOKIE ............ any cookie sent by the client          [yes]
  o  HTTP_FORWARDED ......... list of proxy/gateway hosts            [yes]
  o  HTTP_HOST .............. destination host name/port             [yes]
  o  HTTP_IF_NOT_MODIFIED ... GMT time string                        [yes]
  o  HTTP_PRAGMA ............ any pragma directive of header         [yes]
  o  HTTP_REFERER ........... source document URL for this request   [yes]
  o  HTTP_USER_AGENT ........ client/browser identification string   [yes]

  o  KEY_n .................. query string "+" separated elements    [no]
  o  KEY_COUNT .............. number of "+" separated elements       [no]

  o  PATH_INFO .............. virtual path of data requested in URL  [yes]
  o  PATH_TRANSLATED ........ VMS file path of data requested in URL [yes]

  o  QUERY_STRING ........... string following "?" in URL            [yes]

  o  REMOTE_ADDR ............ IP host address of HTTP client         [yes]
  o  REMOTE_HOST ............ IP host name of HTTP client            [yes]
  o  REMOTE_USER ............ authenticated username (or empty)      [yes]
  o  REQUEST_CHARSET ........ charset SET by mapping rule            [no]
  o  REQUEST_CONTENT_TYPE ... content-type SET by mapping rule       [no]
  o  REQUEST_METHOD ......... "GET", "PUT", etc.                     [yes]
  o  REQUEST_SCHEME ......... "http:" or "https:"                    [no]
  o  REQUEST_TIME_GMT ....... request GMT time                       [no]
  o  REQUEST_TIME_LOCAL ..... request local time                     [no]

  o  SCRIPT_NAME ............ name of script (e.g. "/query")         [yes]
  o  SERVER_ADDR ............ IP host address of server system       [no]
  o  SERVER_CHARSET ......... default charset (e.g. "ISO-8895-1")    [no]
  o  SERVER_GMT ............. offset from GMT time (e.g. "+09:30)    [no]
  o  SERVER_NAME ............ IP host name of server system          [yes]
  o  SERVER_PROTOCOL ........ HTTP protocol version ("HTTP/1.0")     [yes]
  o  SERVER_PORT ............ IP port request was received on        [yes]
  o  SERVER_SOFTWARE ........ software ID of the HTTPD daemon        [yes]

  (these "Purveyor"-like security/SSL variables are available with SSL request)
  o  SECURITY_STATUS ........ "NONE" or "SSL"                        [no]
  o  SSL_CIPHER ............. name of the cipher in use              [no]
  o  SSL_CIPHER_KEYSIZE ..... export:40, domestic: 128               [no]
  o  SSL_CLIENT_CA .......... client certificate authority           [no]
  o  SSL_CLIENT_DN .......... client certificate distinguished name  [no]
  o  SSL_SERVER_CA .......... server certificate authority           [no]
  o  SSL_SERVER_DN .......... server certificate distinguished name  [no]
  o  SSL_VERSION ............ SSL version (e.g. "SSLv2")             [no]

NOTE: If script portability is a concern then confine use to the "standard" CGI
variables.  For WASD-specific scripts the extension variables may be used.


CGI VARIABLES USING DCL SYMBOLS
-------------------------------
Due to the mechanism used to create DCL symbols for the CGI variables (for
standard CGI) the variable value is limited to 1005 characters minus the length
of the DCL symbol name assignment (e.g. "WWW_QUERY_STRING=A+B+C+D"), or in
round terms approximately 980 characters maximum.  This 1005 appears to be a
"magic number" for the DCL command line (VAX VMS V7.2 anyway).


VERSION HISTORY
---------------
17-DEC-1999  MGD  bugfix; quote double-up in CgiVariable()
27-NOV-1999  MGD  rework DCL variable stream,
                  remove sys$fao() from stream CGI variable creation,
                  avoid "''" substitution creating DCL symbol CGI variables
20-NOV-1999  MGD  end-of-header blank line may now be an empty record,
                  or a record containing a single '\n' or '\r\n' sequence
06-NOV-1999  MGD  allow slightly more latitude with form fields, no "=value"
                  (i.e. name terminated by '&') and back-to-back '&'
14-SEP-1999  MGD  modify CgiOutput() for delayed CGI "Content-Type:"
28-AUG-1999  MGD  support CGIplus "agents", added AUTH_USER
30-MAY-1999  MGD  add SSL variables (via call to SesolaCgiGenerateVariables())
21-FEB-1999  MGD  change authorization CGI variables
28-DEC-1998  MGD  added HTTP_ACCEPT_ENCODING
07-NOV-1998  MGD  WATCH facility
17-OCT-1998  MGD  CgiUrlDecodeString(),
                  SERVER_CHARSET, REQUEST_CHARSET, REQUEST_CONTENT_TYPE
28-APR-1998  MGD  CGI variable memory allocation changed in support of SSI
03-APR-1998  MGD  bugfix; extra quotes generated in form field value
19-MAR-1998  MGD  suppress 'Authorization:' unless "external" authorization
06-DEC-1997  MGD  functionality unbundled from DCL.c, generalized for version 5
*/
/*****************************************************************************/

/* standard C header files */
#include <ctype.h>
#include <stdio.h>

/* VMS related header files */
#include <descrip.h>
#include <dvidef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application header files */
#include "wasd.h"

#define WASD_MODULE "CGI"

/***********/
/* defines */
/***********/

#define CLI$_BUFOVF 229400

/********************/
/* external storage */
/********************/

#ifdef DBUG
extern boolean Debug;
#else
#define Debug 0 
#endif

extern int  DclCgiPlusInSize,
            DclSysCommandSize,
            ServerPort,
            WatchEnabled;

extern char  DclCgiVariablePrefix[],
             HttpProtocol[],
             SoftwareID[],
             TimeGmtString[];

extern struct ConfigStruct Config;
extern struct MsgStruct Msgs;

/*****************************************************************************/
/*
Creates a buffer containing a series of null-terminated strings, terminated by
a null-string (two successive null characters).  Each of the strings contains
either DCL command(s) to create a DCL symbol (CGI variable) using the CLI, or
an equate separated 'name=value' pair used for CGIplus variable streams, etc.
*/ 

CgiGenerateVariables
(
struct RequestStruct *rqptr,
int VarType
)
{
   static $DESCRIPTOR (NumberFaoDsc, "!UL\0");
   static $DESCRIPTOR (StatusValueFaoDsc, "%X!8UL\0");

   register char  c;
   register char  *cptr, *sptr, *zptr;

   int  status,
        Count,
        KeyCount;
   char  FormFieldName [256],
         LocalDateTime [48],
         String [1024];
   $DESCRIPTOR (StringDsc, String);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CgiGenerateVariables()\n");

   if (rqptr->WatchItem &&
       (WatchEnabled & WATCH_CGI))
   {
      if (VarType == CGI_VARIABLE_STREAM)
         WatchThis (rqptr, FI_LI, WATCH_CGI, "VARIABLES stream");
      else
         WatchThis (rqptr, FI_LI, WATCH_CGI, "VARIABLES dcl");
   }

   if (rqptr->CgiBufferLength)
   {
      /* free previous variables and storage */
      if (rqptr->CgiBufferPtr != NULL)
         VmFreeFromHeap (rqptr, rqptr->CgiBufferPtr); 
      rqptr->CgiBufferLength = 0;
      rqptr->CgiBufferPtr = rqptr->CgiBufferCurrentPtr = NULL;
   }

   if (VarType == CGI_VARIABLE_STREAM)
      CgiVariableBufferMemory (rqptr, DclCgiPlusInSize*2);
   else
      CgiVariableBufferMemory (rqptr, DclSysCommandSize);

   HttpLocalTimeString (LocalDateTime, &rqptr->BinaryTime);

   if (rqptr->AuthType[0])
   {
      if (rqptr->AuthSourceRealm != NULL)
      { 
         if (VMSnok (status =
             CgiVariable (rqptr, "AUTH_ACCESS", 
                          AuthCanString (rqptr->AuthRequestCan,
                                         AUTH_CAN_FORMAT_LONG),
                          VarType)))
            return (status);
      }

      if (rqptr->AuthAgentParameterPtr != NULL &&
          rqptr->AuthAgentParameterPtr[0])
      { 
         if (VMSnok (status =
             CgiVariable (rqptr, "AUTH_AGENT",
                          rqptr->AuthAgentParameterPtr, VarType)))
            return (status);
      }

      if (rqptr->AuthGroupWritePtr[0])
      {
         if (VMSnok (status =
             CgiVariable (rqptr, "AUTH_GROUP",
                          rqptr->AuthGroupWritePtr, VarType)))
            return (status);
      }
      else
      {
         if (VMSnok (status =
             CgiVariable (rqptr, "AUTH_GROUP",
                          rqptr->AuthGroupReadPtr, VarType)))
            return (status);
      }

      if ((rqptr->AuthAgentParameterPtr != NULL &&
           rqptr->AuthAgentParameterPtr[0]) ||
          rqptr->AuthSourceRealm == AUTH_SOURCE_EXTERNAL)
      {
         if (VMSnok (status =
             CgiVariable (rqptr, "AUTH_PASSWORD",
                          rqptr->RemoteUserPassword, VarType)))
            return (status);
      }

      if (VMSnok (status =
          CgiVariable (rqptr, "AUTH_REALM", rqptr->AuthRealmPtr, VarType)))
         return (status);

      if (VMSnok (status =
          CgiVariable (rqptr, "AUTH_REALM_DESCRIPTION",
                       rqptr->AuthRealmDescriptionPtr, VarType)))
         return (status);
   }

   if (VMSnok (status =
       CgiVariable (rqptr, "AUTH_TYPE", rqptr->AuthType, VarType)))
      return (status);

   if (rqptr->AuthType[0])
   {
      if (VMSnok (status =
          CgiVariable (rqptr, "AUTH_USER", rqptr->AuthUserDetailsPtr, VarType)))
         return (status);
   }

   sys$fao (&NumberFaoDsc, 0, &StringDsc, rqptr->ContentLength);
   if (VMSnok (status =
       CgiVariable (rqptr, "CONTENT_LENGTH", String, VarType)))
      return (status);

   if (rqptr->HttpContentTypePtr != NULL)
   {
      if (VMSnok (status =
          CgiVariable (rqptr, "CONTENT_TYPE",
                       rqptr->HttpContentTypePtr, VarType)))
         return (status);
   }
   else
   if (rqptr->ContentType.ContentTypePtr != NULL)
   {
      if (VMSnok (status =
          CgiVariable (rqptr, "CONTENT_TYPE",
                       rqptr->ContentType.ContentTypePtr, VarType)))
         return (status);
   }
   else
   {
      if (VMSnok (status =
          CgiVariable (rqptr, "CONTENT_TYPE", "", VarType)))
         return (status);
   }

   if (VMSnok (status =
       CgiVariable (rqptr, "GATEWAY_INTERFACE", "CGI/1.1", VarType)))
      return (status);

   if (rqptr->HttpAcceptPtr != NULL)
      if (VMSnok (status =
          CgiVariable (rqptr, "HTTP_ACCEPT", rqptr->HttpAcceptPtr, VarType)))
      return (status);

   if (rqptr->HttpAcceptCharsetPtr != NULL)
      if (VMSnok (status =
          CgiVariable (rqptr, "HTTP_ACCEPT_CHARSET",
                       rqptr->HttpAcceptCharsetPtr, VarType)))
      return (status);

   if (rqptr->HttpAcceptEncodingPtr != NULL)
      if (VMSnok (status =
          CgiVariable (rqptr, "HTTP_ACCEPT_ENCODING",
                       rqptr->HttpAcceptEncodingPtr, VarType)))
      return (status);

   if (rqptr->HttpAcceptLangPtr != NULL)
      if (VMSnok (status =
          CgiVariable (rqptr, "HTTP_ACCEPT_LANGUAGE",
                       rqptr->HttpAcceptLangPtr, VarType)))
         return (status);

   if (rqptr->AuthSourceRealm == AUTH_SOURCE_EXTERNAL &&
       rqptr->HttpAuthorizationPtr != NULL)
      if (VMSnok (status =
          CgiVariable (rqptr, "HTTP_AUTHORIZATION",
                       rqptr->HttpAuthorizationPtr, VarType)))
         return (status);

   if (rqptr->HttpCookiePtr != NULL)
      if (VMSnok (status =
          CgiVariable (rqptr, "HTTP_COOKIE", rqptr->HttpCookiePtr, VarType)))
         return (status);

   if (rqptr->HttpForwardedPtr != NULL)
      if (VMSnok (status =
          CgiVariable (rqptr, "HTTP_FORWARDED",
                       rqptr->HttpForwardedPtr, VarType)))
         return (status);

   if (rqptr->HttpHostPtr != NULL)
      if (VMSnok (status =
          CgiVariable (rqptr, "HTTP_HOST", rqptr->HttpHostPtr, VarType)))
         return (status);

   if (rqptr->HttpIfModifiedSincePtr != NULL)
      if (VMSnok (status =
          CgiVariable (rqptr, "HTTP_IF_MODIFIED_SINCE",
                       rqptr->HttpIfModifiedSincePtr, VarType)))
         return (status);

   if (rqptr->HttpPragmaPtr != NULL)
      if (VMSnok (status =
          CgiVariable (rqptr, "HTTP_PRAGMA", rqptr->HttpPragmaPtr, VarType)))
         return (status);

   if (rqptr->HttpRefererPtr != NULL)
      if (VMSnok (status =
          CgiVariable (rqptr, "HTTP_REFERER", rqptr->HttpRefererPtr, VarType)))
         return (status);

   if (rqptr->HttpUserAgentPtr != NULL)
      if (VMSnok (status =
          CgiVariable (rqptr, "HTTP_USER_AGENT",
                       rqptr->HttpUserAgentPtr, VarType)))
         return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "PATH_INFO", rqptr->PathInfoPtr, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "PATH_TRANSLATED", rqptr->RequestFileName, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "QUERY_STRING", rqptr->QueryStringPtr, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "REMOTE_ADDR",
                    rqptr->ClientIpAddressString, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "REMOTE_HOST", rqptr->ClientHostName, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "REMOTE_USER", rqptr->RemoteUser, VarType)))
      return (status);

   if (rqptr->SetPathCharsetPtr != NULL)
   {
      if (VMSnok (status =
          CgiVariable (rqptr, "REQUEST_CHARSET",
                       rqptr->SetPathCharsetPtr, VarType)))
         return (status);
   }

   if (rqptr->SetPathContentTypePtr != NULL)
   {
      if (VMSnok (status =
          CgiVariable (rqptr, "REQUEST_CONTENT_TYPE",
                       rqptr->SetPathContentTypePtr, VarType)))
         return (status);
   }

   if (VMSnok (status =
       CgiVariable (rqptr, "REQUEST_METHOD", rqptr->HttpMethodName, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "REQUEST_SCHEME",
                    rqptr->RequestSchemeNamePtr, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "REQUEST_TIME_GMT", rqptr->GmDateTime, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "REQUEST_TIME_LOCAL", LocalDateTime, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "SCRIPT_NAME", rqptr->ScriptName, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "SERVER_CHARSET", Config.CharsetDefault, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "SERVER_GMT", TimeGmtString, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "SERVER_NAME",
                    rqptr->ServicePtr->ServerHostName, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "SERVER_ADDR",
                    rqptr->ServicePtr->ServerIpAddressString, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "SERVER_PROTOCOL", HttpProtocol, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "SERVER_PORT",
                    rqptr->ServicePtr->ServerPortString, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "SERVER_SOFTWARE", SoftwareID, VarType)))
      return (status);

   /*****************/
   /* SSL variables */
   /*****************/

   if (VMSnok (status = SesolaCgiGenerateVariables (rqptr, VarType)))
      return (status);

   /***************************/
   /* query string components */
   /***************************/

   if (rqptr->QueryStringPtr[0])
   {
      KeyCount = 0;

      cptr = rqptr->QueryStringPtr;
      while (*cptr && *cptr != '=') cptr++;
      /* if an equal symbol was found then its a form not a keyword search */
      if (*cptr)
      {
         /***************/
         /* form fields */
         /***************/

         memcpy (FormFieldName, "FORM_", 5);
         cptr = rqptr->QueryStringPtr;
         while (*cptr)
         {
            sptr = FormFieldName + 5;
            zptr = FormFieldName + sizeof(FormFieldName);
            while (*cptr && *cptr != '=' && *cptr != '&' && sptr < zptr)
            {
               if (isalnum(*cptr))
                  *sptr++ = toupper(*cptr++);
               else
               {
                  *sptr++ = '_';
                  cptr++;
               }
            }
            if (sptr >= zptr)
            {
               ErrorVmsStatus (rqptr, SS$_BUFFEROVF, FI_LI);
               return (STS$K_ERROR);
            }
            *sptr = '\0';

            if (!FormFieldName[5] || (*cptr && *cptr != '=' && *cptr != '&'))
            {
               if (*cptr == '&')
               {
                  /* allow back-to-back '&' (effectively null field name) */
                  cptr++;
                  continue;
               }
               /* error, out-of-place '=' (no field name) */
               rqptr->ResponseStatusCode = 400;
               ErrorGeneral (rqptr, MsgFor(rqptr,MSG_REQUEST_URL_FORM), FI_LI);
               return (STS$K_ERROR);
            }

            /* if encountered an '=' */
            if (*cptr == '=') cptr++;
            Count = CgiUrlDecodeString (rqptr, cptr,
                                        String, sizeof(String), '&');
            if (Count < 0) return (STS$K_ERROR);
            cptr += Count;
            if (VMSnok (status =
                CgiVariable (rqptr, FormFieldName, String, VarType)))
               return (status);
         }

         if (VMSnok (status =
             CgiVariable (rqptr, "KEY_COUNT", "0", VarType)))
            return (status);
      }
      else
      {
         /******************/
         /* query keywords */
         /******************/

         cptr = rqptr->QueryStringPtr;
         while (*cptr)
         {
            sys$fao (&NumberFaoDsc, 0, &StringDsc, ++KeyCount);
            memcpy (FormFieldName, "KEY_", 4);
            strcpy (FormFieldName+4, String);

            Count = CgiUrlDecodeString (rqptr, cptr,
                                        String, sizeof(String), '+');
            if (Count < 0) return (STS$K_ERROR);
            cptr += Count;
            if (VMSnok (status =
                CgiVariable (rqptr, FormFieldName, String, VarType)))
               return (status);
         }

         sys$fao (&NumberFaoDsc, 0, &StringDsc, KeyCount);
         if (VMSnok (status =
             CgiVariable (rqptr, "KEY_COUNT", String, VarType)))
            return (status);
      }
   }
   else
   {
      /* no keywords if no query string! */
      if (VMSnok (status =
          CgiVariable (rqptr, "KEY_COUNT", "0", VarType)))
         return (status);
   }

   return (CgiVariable (rqptr, NULL, NULL, VarType));
}

/*****************************************************************************/
/*
URL-decode a URL-encoded string!  Return -1 if an error occurs, otherwise
return the number of characters read from the original, encoded string.
*/ 

int CgiUrlDecodeString
(
struct RequestStruct *rqptr,
char *EncodedString,
char *DecodedString,
int SizeofDecodedString,
char Separator
)
{
   register char  c;
   register char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CgiUrlDecodeString() |%s|\n", EncodedString);

   cptr = EncodedString;
   zptr = (sptr = DecodedString) + SizeofDecodedString;
   while (*cptr && *cptr != Separator && sptr < zptr)
   {
      if (*cptr == '+')
      {
         *sptr++ = ' ';
         cptr++;
      }
      else
      if (*cptr == '%')
      {
         /* an escaped character ("%xx" where xx is a hex number) */
         cptr++;
         c = 0;
         if (*cptr >= '0' && *cptr <= '9')
            { c = (*cptr - (int)'0') << 4; cptr++; }
         else
         if (tolower(*cptr) >= 'a' && tolower(*cptr) <= 'f')
            { c = (tolower(*cptr) - (int)'a' + 10) << 4; cptr++; }
         else
         {
            rqptr->ResponseStatusCode = 400;
            ErrorGeneral (rqptr, MsgFor(rqptr,MSG_REQUEST_URL_ENC), FI_LI);
            return (-1);
         }
         if (*cptr >= '0' && *cptr <= '9')
            { c += (*cptr - (int)'0'); cptr++; }
         else
         if (tolower(*cptr) >= 'a' && tolower(*cptr) <= 'f')
            { c += (tolower(*cptr) - (int)'a' + 10); cptr++; }
         else
         {
            rqptr->ResponseStatusCode = 400;
            ErrorGeneral (rqptr, MsgFor(rqptr,MSG_REQUEST_URL_ENC), FI_LI);
            return (-1);
         }
         if (sptr < zptr) *sptr++ = c;
      }
      else
         *sptr++ = *cptr++;
   }
   if (sptr >= zptr)
   {
      ErrorGeneral (rqptr, MsgFor(rqptr,MSG_GENERAL_OVERFLOW), FI_LI);
      return (-1);
   }
   *sptr = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", DecodedString);
   if (*cptr) cptr++;

   return (cptr - EncodedString);
}

/*****************************************************************************/
/*
Depending on whether the CGI variable is for creation at the DCL CLI or as part
of a CGI variable stream (e.g. CGIplus) create a null-terminated string
conatining either commands to create a DCL symbol, or a 'name=value' pair.

DCL symbol creation at the command line is limited by the CLI command line 
length (255 characters).  Symbol values however can be up to approximately 
1000 characters, probably enough for any CGI variable value.  If a CGI value 
is too large for for a single command-line assignment then build it up using 
multiple assignments, a symbol assignment kludge!
*/ 

int CgiVariable
(
struct RequestStruct *rqptr,
char *SymbolName,
char *SymbolValue,
int VarType
)
{
   static char  DeleteAllLocalSymbol[] = "DELETE/SYMBOL/LOCAL/ALL";

   register char  *cptr, *sptr, *zptr;

   int  status,
        SymbolCount,
        SymbolValueLength;
   unsigned short  Length;
   char  *CgiVariablePrefixPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "CgiVariable() |%s|%s|\n", SymbolName, SymbolValue);

   if (SymbolName == NULL)
   {
      /************************/
      /* end of CGI variables */
      /************************/

      if (rqptr->CgiBufferRemaining < sizeof(short))
         CgiVariableBufferMemory (rqptr, sizeof(short));

      /* zero length record terminates generated CGI variables */
      *(short*)rqptr->CgiBufferCurrentPtr = 0;
      rqptr->CgiBufferRemaining -= sizeof(short);
      rqptr->CgiBufferCurrentPtr += sizeof(short);

      if (Debug)
         fprintf (stdout, "buffer used: %d\n",
                  rqptr->CgiBufferCurrentPtr - rqptr->CgiBufferPtr);

      return (SS$_NORMAL);
   }

   if (rqptr->SetPathCgiPrefixPtr == NULL)
      CgiVariablePrefixPtr = DclCgiVariablePrefix;
   else
      CgiVariablePrefixPtr = rqptr->SetPathCgiPrefixPtr;

   if (SymbolValue == NULL) SymbolValue = "";

   if (rqptr->WatchItem &&
       (WatchEnabled & WATCH_CGI))
      WatchDataFormatted ("!AZ=!AZ\n", SymbolName, SymbolValue);

   if (VarType == CGI_VARIABLE_STREAM)
   {
      /******************************/
      /* variables in record stream */
      /******************************/

      if (rqptr->CgiBufferRemaining <= DclCgiPlusInSize + sizeof(short))
         CgiVariableBufferMemory (rqptr, DclCgiPlusInSize*2);

      zptr = (sptr = rqptr->CgiBufferCurrentPtr + sizeof(short)) +
             (rqptr->CgiBufferRemaining - sizeof(short));
      for (cptr = CgiVariablePrefixPtr; *cptr; *sptr++ = *cptr++);
      for (cptr = SymbolName; *cptr && sptr < zptr; *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = '=';
      for (cptr = SymbolValue; *cptr && sptr < zptr; *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = '\0';
      if (sptr < zptr)
      {
         Length = sptr - (char*)rqptr->CgiBufferCurrentPtr - sizeof(short);
         *(short*)rqptr->CgiBufferCurrentPtr = Length;
         rqptr->CgiBufferRemaining -= Length + sizeof(short);
         rqptr->CgiBufferCurrentPtr += Length + sizeof(short);
         return (SS$_NORMAL);
      }
      ErrorGeneral (rqptr, MsgFor(rqptr,MSG_GENERAL_OVERFLOW), FI_LI);
      return (STS$K_ERROR);
   }

   /********************************/
   /* variables created at DCL CLI */
   /********************************/

   if (rqptr->CgiBufferRemaining <= 256 + sizeof(short))
      CgiVariableBufferMemory (rqptr, DclSysCommandSize);

   zptr = (sptr = rqptr->CgiBufferCurrentPtr + sizeof(short)) + 255;
   for (cptr = CgiVariablePrefixPtr; *cptr; *sptr++ = *cptr++);
   for (cptr = SymbolName; *cptr && sptr < zptr; *sptr++ = *cptr++);
   if (sptr < zptr) *sptr++ = '=';
   if (sptr < zptr) *sptr++ = '=';
   if (sptr < zptr) *sptr++ = '\"';
   cptr = SymbolValue;
   while (*cptr && sptr < zptr)
   {
      if (*cptr == '\"')
      {
         /* escape double quotes, by doubling up */
         if (sptr+2 < zptr)
         {
            cptr++;
            *sptr++ = '\"';
            *sptr++ = '\"';
         }
         else
            zptr = sptr;
         continue;
      }

      if (*cptr == '\'')
      {
         /* escape single quotes by concatenation */
         if (sptr+7 < zptr)
         {
            cptr++;
            memcpy (sptr, "\"+\"\'\"+\"", 7);
            sptr += 7;
         }
         else
            sptr = zptr;
         continue;
      }

      if (sptr < zptr) *sptr++ = *cptr++;
   }
   if (sptr < zptr) *sptr++ = '\"';
   if (sptr < zptr)
   {
      *sptr++ = '\0';
      Length = sptr - (char*)rqptr->CgiBufferCurrentPtr - sizeof(short);
      *(short*)rqptr->CgiBufferCurrentPtr = Length;
      rqptr->CgiBufferRemaining -= Length + sizeof(short);
      rqptr->CgiBufferCurrentPtr += Length + sizeof(short);
      return (SS$_NORMAL);
   }

   /* loop assigning maximum amount allowed by DCL until all assigned */
   SymbolCount = 0;
   cptr = SymbolValue;
   while (*cptr)
   {
      if (rqptr->CgiBufferRemaining <= 256 + sizeof(short))
         CgiVariableBufferMemory (rqptr, DclSysCommandSize);

      if (Debug) fprintf (stdout, "|%s|\n", cptr);

      /* set this to DCL line length minus one for the terminating quote */
      zptr = (sptr = rqptr->CgiBufferCurrentPtr + sizeof(short)) + 254;
      switch (++SymbolCount)
      {
         case 1 : memcpy (sptr, "A=\"", 3); break;
         case 2 : memcpy (sptr, "B=\"", 3); break;
         case 3 : memcpy (sptr, "C=\"", 3); break;
         case 4 : memcpy (sptr, "D=\"", 3); break;
         case 5 : memcpy (sptr, "E=\"", 3); break;
         case 6 : memcpy (sptr, "F=\"", 3); break;
         default : memcpy (sptr, "X=\"", 3); break;
      }
      sptr += 3;
      while (*cptr && sptr < zptr)
      {
         if (*cptr == '\"')
         {
            /* escape double quotes, by doubling up */
            if (sptr+2 < zptr)
            {
               cptr++;
               *sptr++ = '\"';
               *sptr++ = '\"';
            }
            else
               zptr = sptr;
            continue;
         }

         if (*cptr == '\'')
         {
            /* escape single quotes by concatenation */
            if (sptr+7 < zptr)
            {
               cptr++;
               memcpy (sptr, "\"+\"\'\"+\"", 7);
               sptr += 7;
            }
            else
               sptr = zptr;
            continue;
         }

         if (sptr < zptr) *sptr++ = *cptr++;
      }

      *sptr++ = '\"';
      *sptr++ = '\0';
      Length = sptr - (char*)rqptr->CgiBufferCurrentPtr - sizeof(short);
      *(short*)rqptr->CgiBufferCurrentPtr = Length;
      rqptr->CgiBufferRemaining -= Length + sizeof(short);
      rqptr->CgiBufferCurrentPtr += Length + sizeof(short);
   }
   SymbolValueLength = cptr - SymbolValue;

   /* assign the temporary symbol value(s) to the CGI symbol */
   if (rqptr->CgiBufferRemaining <= 256 + sizeof(short))
      CgiVariableBufferMemory (rqptr, DclSysCommandSize);
   sptr = rqptr->CgiBufferCurrentPtr + sizeof(short);
   for (cptr = CgiVariablePrefixPtr; *cptr; *sptr++ = *cptr++);
   for (cptr = SymbolName; *cptr; *sptr++ = *cptr++);
   switch (SymbolCount)
   {
      case 1 : memcpy (sptr, "==A", 4); sptr += 4; break;
      case 2 : memcpy (sptr, "==A+B", 6); sptr += 6; break;
      case 3 : memcpy (sptr, "==A+B+C", 8); sptr += 8; break;
      case 4 : memcpy (sptr, "==A+B+C+D", 10); sptr += 10; break;
      case 5 : memcpy (sptr, "==A+B+C+D+E", 12); sptr += 12; break;
      case 6 : memcpy (sptr, "==A+B+C+D+E+F", 14); sptr += 14; break;
      default : break;
   }
   Length = sptr - (char*)rqptr->CgiBufferCurrentPtr - sizeof(short);

   /* 1005 seems to be a MAGIC TOTAL LINE LENGTH (for this approach anyway) */
   if (SymbolCount > 6 ||
       Length + SymbolValueLength > 1006)
   {
      /* cannot create a symbol this size */
      rqptr->ErrorTextPtr = SymbolName;
      rqptr->ErrorOtherTextPtr = "DCL symbol assignment too large!";
      ErrorVmsStatus (rqptr, CLI$_BUFOVF, FI_LI);
      return (STS$K_ERROR);
   }

   *(short*)rqptr->CgiBufferCurrentPtr = Length;
   rqptr->CgiBufferRemaining -= Length + sizeof(short);
   rqptr->CgiBufferCurrentPtr += Length + sizeof(short);

   /* not really necessary, but let's be tidy */
   if (rqptr->CgiBufferRemaining <= sizeof(DeleteAllLocalSymbol) +
                                    sizeof(short))
      CgiVariableBufferMemory (rqptr, DclSysCommandSize);
   memcpy (rqptr->CgiBufferCurrentPtr + sizeof(short),
           DeleteAllLocalSymbol, sizeof(DeleteAllLocalSymbol));
   *(short*)rqptr->CgiBufferCurrentPtr = sizeof(DeleteAllLocalSymbol);
   rqptr->CgiBufferRemaining -= sizeof(DeleteAllLocalSymbol) + sizeof(short);
   rqptr->CgiBufferCurrentPtr += sizeof(DeleteAllLocalSymbol) + sizeof(short);

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Allocate, or reallocate on subsequent calls, memory for storing the CGI
variables in.  Adjusts the CGI variable buffer pointers, lengths, etc,
appropriately.
*/ 

CgiVariableBufferMemory
(
struct RequestStruct *rqptr,
int BufferChunk
)
{
   int  CurrentLength,
        CurrentOffset;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "CgiVariableBufferMemory() %d %d\n",
               rqptr->CgiBufferLength, BufferChunk);

   if (!rqptr->CgiBufferLength)
   {
      /* initialize CGI variable buffer */
      rqptr->CgiBufferLength = rqptr->CgiBufferRemaining = BufferChunk;
      rqptr->CgiBufferPtr = rqptr->CgiBufferCurrentPtr =
         VmGetHeap (rqptr, BufferChunk);
      return;
   }

   CurrentLength = rqptr->CgiBufferLength;
   CurrentOffset = rqptr->CgiBufferCurrentPtr - rqptr->CgiBufferPtr;

   rqptr->CgiBufferPtr =
      VmReallocHeap (rqptr, rqptr->CgiBufferPtr, CurrentLength + BufferChunk);

   rqptr->CgiBufferLength = CurrentLength + BufferChunk;
   rqptr->CgiBufferRemaining += BufferChunk;
   rqptr->CgiBufferCurrentPtr = rqptr->CgiBufferPtr + CurrentOffset;
}

/*****************************************************************************/
/*
Process output from a CGI script (or even plain DCL processing).  For CGI
scripts check the first output from each for CGI-relevant HTTP header lines. 
For all output check that carriage-control is as required.
*/ 

int CgiOutput
(
struct RequestStruct *rqptr,
char *OutputPtr,
int OutputCount
)
{
   int  value;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CgiOutput() %d\n", OutputCount);

   if (rqptr->SequenceEofLength)
   {
      if (*(long*)OutputPtr == *(long*)rqptr->SequenceEofPtr &&
          OutputCount >= rqptr->SequenceEofLength &&
          OutputCount <= rqptr->SequenceEofLength+2 &&
          !memcmp (OutputPtr, rqptr->SequenceEofPtr, rqptr->SequenceEofLength))
      {
         /******************************/
         /* end of output from script! */
         /******************************/

         if (Debug) fprintf (stdout, "OUTPUT EOF!\n");
         return (CGI_OUTPUT_END);
      }
   }

   if (rqptr->CgiPlusCalloutInProgress)
   {
      /********************************/
      /* output escape is in progress */
      /********************************/

      if (*(long*)OutputPtr == *(long*)rqptr->SequenceEotPtr &&
          OutputCount >= rqptr->SequenceEotLength &&
          OutputCount <= rqptr->SequenceEotLength+2 &&
          !memcmp (OutputPtr, rqptr->SequenceEotPtr, rqptr->SequenceEotLength))
      {
         /******************************/
         /* end of escape from script! */
         /******************************/

         if (Debug) fprintf (stdout, "OUTPUT EOT!\n");
         return (CGI_OUTPUT_ESCAPE_END);
      }

      return (CGI_OUTPUT_ESCAPE);
   }

   if (rqptr->SequenceEscLength)
   {
      if (*(long*)OutputPtr == *(long*)rqptr->SequenceEscPtr &&
          OutputCount >= rqptr->SequenceEscLength &&
          OutputCount <= rqptr->SequenceEscLength+2 &&
          !memcmp (OutputPtr, rqptr->SequenceEscPtr, rqptr->SequenceEscLength))
      {
         /******************************/
         /* escape from script output! */
         /******************************/

         if (Debug) fprintf (stdout, "OUTPUT ESC!\n");
         return (CGI_OUTPUT_ESCAPE_BEGIN);
      }
   }

   if (rqptr->LocationPtr != NULL)
   {
      /* absorb any output from script until the request processing ends */
      return (CGI_OUTPUT_ABSORB);
   }

   if (rqptr->DclTaskPtr != NULL &&
       (rqptr->DclTaskPtr->TaskType == DCL_TASK_TYPE_CGI_SCRIPT ||
        rqptr->DclTaskPtr->TaskType == DCL_TASK_TYPE_CGIPLUS_SCRIPT) ||
       rqptr->DECnetTaskPtr != NULL)
   {
      /********************/
      /* script execution */
      /********************/

      if (!rqptr->CgiCheckedHeader)
      {
         value = CgiScriptHeader (rqptr, OutputPtr, OutputCount);
         if (value == CGI_OUTPUT_END ||
             value == CGI_OUTPUT_ABSORB)
            return (value);
      }
   }

   if (rqptr->CgiAddNewline)
   {
      /*************************/
      /* HTTP carriage-control */
      /*************************/

      if (Debug) fprintf (stdout, "rqptr->CgiAddNewline\n");

      if (OutputCount)
      {
         if (OutputPtr[OutputCount-1] != '\n')
         {
            OutputPtr[OutputCount++] = '\n';
            OutputPtr[OutputCount] = '\0';
         }

         if (rqptr->CgiInHeader)
         {
            /* still processing records from the response header */
            if ((OutputCount == 1 && OutputPtr[0] == '\n') ||
                (OutputCount == 2 && *(unsigned short*)OutputPtr == '\r\n'))
            {
               /* first blank line is end of response header! */
               rqptr->CgiInHeader = false;
               if (!rqptr->CgiContentTypeText) rqptr->CgiAddNewline = false;
            }
            else
            if (toupper(OutputPtr[0]) == 'C' &&
                strsame (OutputPtr, "Content-Type:", 13))
            {
               /* content-type specified */
               char  *cptr = OutputPtr + 13;
               while (ISLWS(*cptr) && NOTEOL(*cptr)) cptr++;
               if (strsame (cptr, "text/", 5))
               {
                  if (Debug) fprintf (stdout, "text/\n");
                  rqptr->CgiContentTypeText = true;
               }
            }
         }
      }
      else
      {
         /* must be a blank line (empty record), add a newline */
         OutputPtr[OutputCount++] = '\n';
         OutputPtr[OutputCount] = '\0';

         /* first blank line is end of response header! */
         if (rqptr->CgiInHeader)
         {
            rqptr->CgiInHeader = false;
            if (!rqptr->CgiContentTypeText) rqptr->CgiAddNewline = false;
         }
      }
   }

   return (OutputCount);
}

/*****************************************************************************/
/*
For CGI-compliance the first line of the header should be a "Content-Type:", 
"Location:" or "Status:".  If a "Content-Type:" prepend a "200" header status 
line.  If a "Status:" use this to create a unique header status line.  If a 
"Location:" redirection header line call a function to establish the level of 
redirection, and do any associated processing.  For non-parse-header style CGI 
scripts check if the first line is a full HTTP status line.  If it is then do 
nothing, the script will supply the raw HTTP stream.  If none of the above 
then consider the script will supply nothing but plain text, none of the HTTP 
header, etc., and prepend a "200" status line and a "Content-Type: text/plain" 
line to create a complete HTTP header.  This is an extension to CGI behaviour, 
but makes simple, plain-text DCL scripts very easy!  Return true any action
required has been taken by this function, false if further processing is
required by the calling function.

Returns -1 to terminate processing, 0 to continue but absorb any output
(redirection), and 1 to continue processing.
*/ 

int CgiScriptHeader 
(
struct RequestStruct *rqptr,
char *OutputPtr,
int OutputCount
)
{
   register char  *cptr, *sptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CgiScriptHeader()\n|%s|\n", OutputPtr);

   rqptr->CgiCheckedHeader = rqptr->CgiInHeader = true;
   cptr = OutputPtr;

   if (toupper(*cptr) == 'H' && strsame (cptr, "HTTP/", 5))
   {
      /*************************************************/
      /* script is supplying the full HTTP data stream */
      /*************************************************/

      if (Debug) fprintf (stdout, "(HTTP data stream)\n");
      rqptr->CgiAddNewline = false;

      /* a simple way to generate 0.9 responses, just for testing purposes! */
      if (!memcmp (cptr+4, "/0.9", 4)) *(long*)cptr = 'PTTH';

      /* get the response status code for logging purposes */
      while (*cptr && !ISLWS(*cptr)) cptr++;
      while (*cptr && !isdigit(*cptr)) cptr++;
      if (isdigit(*cptr)) rqptr->ResponseStatusCode = atoi(cptr);
      return (OutputCount);
   }
   else
   if (toupper(*cptr) == 'L' && strsame (cptr, "Location:", 9))
   {
      /*******************/
      /* redirection URL */
      /*******************/

      cptr += 9;
      /* locate the first character of the "Location:" URL */
      while (ISLWS(*cptr)) cptr++;
      sptr = cptr;
      while (*cptr && !ISLWS(*cptr) && *cptr != '\r' && *cptr != '\n') cptr++;
      rqptr->LocationPtr = VmGetHeap (rqptr, cptr-sptr+1);
      memcpy (rqptr->LocationPtr, sptr, cptr-sptr);
      rqptr->LocationPtr[cptr-sptr] = '\0';
      if (Debug) fprintf (stdout, "LocationPtr |%s|\n", rqptr->LocationPtr);
      return (CGI_OUTPUT_ABSORB);
   }
   else
   if (toupper(*cptr) == 'C' && strsame (cptr, "Content-Type:", 13))
   {

      /**************************/
      /* content-type specified */
      /**************************/

      /*
         For CGI content types of "text/..." add carriage control to each
         record needing it.  For non-text types this carriage-control is
         added to response header lines then turned off for the body.
      */

      cptr += 13;
      while (ISLWS(*cptr) && NOTEOL(*cptr)) cptr++;
      if (strsame (cptr, "text/", 5))
      {
         /* make sure each record received has HTTP carriage control */
         if (Debug) fprintf (stdout, "(RMS record stream)\n");
         rqptr->CgiAddNewline = rqptr->CgiContentTypeText = true;
      }
      else
         rqptr->CgiAddNewline = true;

      rqptr->ResponseStatusCode = 200;
      HttpHeader (rqptr, rqptr->ResponseStatusCode, NULL, -1, NULL, NULL);

      /* remove generated empty header line to allow the script to supply it */
      if (rqptr->ResponseHeaderLength > 2)
      {
         rqptr->ResponseHeaderPtr[rqptr->ResponseHeaderLength-2] = '\0';
         rqptr->ResponseHeaderLength -= 2;
      }
      return (OutputCount);
   }
   else
   if (toupper(*cptr) == 'S' && strsame (cptr, "Status:", 7))
   {
      /********************/
      /* status specified */
      /********************/

      /* make sure each record received has HTTP carriage control */
      if (Debug) fprintf (stdout, "(RMS record stream)\n");
      rqptr->CgiAddNewline = true;

      /* create HTTP header status line using supplied status */
      cptr += 7;
      while (*cptr++ && !isdigit(*cptr) && *cptr != '\n') cptr++;
      /* get the response status code for header and logging purposes */
      if (isdigit(*cptr)) rqptr->ResponseStatusCode = atoi(cptr);

      HttpHeader (rqptr, rqptr->ResponseStatusCode,
                  "text/plain", -1, NULL, NULL);

      /* remove generated empty header line, allow the script to supply it */
      if (rqptr->ResponseHeaderLength > 2)
      {
         rqptr->ResponseHeaderPtr[rqptr->ResponseHeaderLength-2] = '\0';
         rqptr->ResponseHeaderLength -= 2;
      }
      return (OutputCount);
   }
   else
   {
      /******************/
      /* non-CGI output */
      /******************/

      /* make sure each record received has HTTP carriage control */
      if (Debug) fprintf (stdout, "(RMS record stream)\n");
      rqptr->CgiAddNewline = rqptr->CgiContentTypeText = true;
      rqptr->ResponseStatusCode = 200;

      HttpHeader (rqptr, rqptr->ResponseStatusCode,
                  "text/plain", -1, NULL, NULL);

      return (OutputCount);
   }
}

/*****************************************************************************/
/*
Generate a (hopefully) unique sequence of 224 bits (28 bytes, which is a fair
bit :^) The string is generated using the a count sequence and quadword binary
times each time it is called.  This provides a continuous, non-repeating series
of unlikely bit combinations with a one in 2^224 chance (I think!) of presence
in an I/O stream.
*/ 

CgiSequence
(
char OneChar,
char *CgiSequencePtr,
int *CgiSequenceLengthPtr
)
{                                           
   static $DESCRIPTOR (SequenceFaoDsc, "^!AZ-!8XL!8XL!8XL-\0");
   static $DESCRIPTOR (SequenceDsc, "");
   static char  OneCharString [] = " ";
   static int  SequenceCount = 0;

   unsigned short  Length;
   unsigned long  BinTime[2];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CgiSequence()\n");

   SequenceCount++;
   sys$gettim(&BinTime);
   SequenceDsc.dsc$w_length = 29;
   SequenceDsc.dsc$a_pointer = CgiSequencePtr;
   OneCharString[0] = OneChar;
   sys$fao (&SequenceFaoDsc, &Length, &SequenceDsc,
            OneCharString, SequenceCount, BinTime[1], BinTime[0]);
   *CgiSequenceLengthPtr = Length - 1;

   if (Debug)
      fprintf (stdout, "%d |%s|\n", *CgiSequenceLengthPtr, CgiSequencePtr);
}

/*****************************************************************************/

